@@ -36,6 +36,50 @@
 //! - Recursive parsers
 //! - Text-specific parsers & utilities
 //! - Custom error types
+//!
+//! ## What is LL(k) parsing?
+//!
+//! We can think of parsing like navigating a maze that has no cycles. As we walk towards what we hope is the exit we
+//! will, from time to time, come across splits in the path. Our job is to decide which path to take in order to reach
+//! the exit. How do we do this?
+//!
+//! Thankfully, we have signposts at every turning in the form of upcoming tokens to parse. Unfortunately, these
+//! signposts don't tell us exactly where to go: they just give us an idea of what obstacles might be lying ahead. We
+//! also have a way to see into the future using 'token lookahead'. Lookahead means that at every split in the path, we
+//! get a chance to peer into the distance to see what lies ahead on each possible path before deciding to commit to
+//! it. The distance that we're allowed to look is the 'k' in LL(k) parsing. Parsers capable of taking a walk into the
+//! future, realising they've taken a wrong turn, and retracing their steps are known as 'backtracking' parsers for
+//! this reason.
+//!
+//! ## LL(1) parsing: limitations and advantages
+//!
+//! Chumsky is designed for LL(1) parsing. This means that at each split in the path, we only get to see the very next
+//! token in advance before making a choice about which path to proceed down. In effect, we need to be able to figure
+//! out how to proceed using only the very next token in the input sequence. No backtracking allowed!
+//!
+//! This may initially seem like a serious limitation and it is, at least in theory. In practice, we can eliminate many
+//! of these limitations by deploying *parsing passes*. A common example of multi-pass parsing is the way in which
+//! compilers for C-like languages generally perform an initial lexing (also called 'tokenization') pass to split
+//! the source code into logical groups of characters known as 'tokens' and later feed these tokens into a secondary
+//! pass to generate the program's abstract syntax tree. This isn't just a workaround for the limitations of an LL(1)
+//! parser either: splitting parsing into multiple passes makes writing and validating each pass simpler and allows for
+//! tools like syntax highlighters that make use of earlier passes only to be built on top of the same codebase.
+//!
+//! All of this considered, LL(1) parsers are powerful tools that continue to be a relevant and effective route for
+//! parsing even (and, perhaps, *in particular*) the latest generations of programming languages.
+//!
+//! Why use LL(1) parsing when more powerful backtracking alternatives exist? The answer, in my view, comes down to
+//! three main points.
+//!
+//! - **Simplicity**: LL(1) grammars are usually easier to write (and, arguably, read) than more complex grammars.
+//!
+//! - **Performance**: LL(1) parsing does not require backtracking (because backtracking is impossible). This means
+//!   that parsing can only ever proceed in the forward direction, meaning that parsing has *linear* (i.e: **O(n)**)
+//!   time complexity. Compare this to other parsing techniques, for which exponential time is virtually the norm.
+//!
+//! - **Error recovery**: Because parsers of LL(1) grammars do not perform backtracking, there are fewer potential
+//!   parse paths to consider: this makes it more likely that errors can be recovered from without misinterpreting
+//!   syntax and makes it more likely that errors relate to the syntax that the user is attempting to write.
 
 #![deny(missing_docs)]
 // TODO: Enable when stable
@@ -45,6 +89,8 @@
 pub mod combinator;
 /// Error types, traits and utilities.
 pub mod error;
+/// Traits that allow chaining parser outputs together.
+pub mod chain;
 /// Parser primitives that accept specific token patterns.
 pub mod primitive;
 /// Recursive parsers (parser that include themselves within their patterns).
@@ -54,9 +100,11 @@ pub mod stream;
 /// Text-specific parsers and utilities.
 pub mod text;
 
+pub use crate::error::{Error, Span};
+
 use crate::{
+    chain::Chain,
     combinator::*,
-    error::Error,
     primitive::*,
     stream::*,
 };
@@ -64,7 +112,6 @@ use crate::{
 use std::{
     iter::Peekable,
     marker::PhantomData,
-    rc::Rc,
     // TODO: Enable when stable
     //lazy::OnceCell,
 };
@@ -74,20 +121,19 @@ pub mod prelude {
     pub use super::{
         error::{Error as _, Simple},
         text::{TextParser as _, whitespace},
-        primitive::{just, filter, end, any},
+        primitive::{any, end, filter, filter_map, just, one_of, seq},
         recursive::recursive,
+        text,
         Parser,
     };
 }
 
-/*
 fn or_zip_with<T, F: FnOnce(T, T) -> T>(a: Option<T>, b: Option<T>, f: F) -> Option<T> {
     match (a, b) {
         (Some(a), Some(b)) => Some(f(a, b)),
         (a, b) => a.or(b),
     }
 }
-*/
 
 fn zip_or<T, F: FnOnce(T, T) -> T>(a: Option<T>, b: T, f: F) -> T {
     match a {
@@ -115,21 +161,21 @@ fn zip_or<T, F: FnOnce(T, T) -> T>(a: Option<T>, b: T, f: F) -> T {
 /// have a lexical pass prior to the main parser that groups the input characters into tokens.
 pub trait Parser<I, O> {
     /// The type of errors emitted by this parser.
-    type Error: Error<I>;
+    type Error: Error<I>; // TODO when default associated types are stable: = Simple<I>;
 
     /// Parse a stream with all the bells & whistles. You can use this to implement your own parser combinators. Note
     /// that both the signature and semantic requirements of this function are very likely to change in later versions.
     /// Where possible, prefer more ergonomic combinators provided elsewhere in the crate rather than implementing your
     /// own.
-    fn parse_inner<S: Stream<I>>(&self, stream: &mut S, errors: &mut Vec<Self::Error>) -> (usize, Result<O, Self::Error>) where Self: Sized;
+    fn parse_inner<S: Stream<I, <Self::Error as Error<I>>::Span>>(&self, stream: &mut S, errors: &mut Vec<Self::Error>) -> (usize, Result<(O, Option<Self::Error>), Self::Error>) where Self: Sized;
 
     /// Parse an iterator of tokens, yielding an output if possible, and any errors encountered along the way.
     ///
     /// If you don't care about producing an output if errors are encountered, use `Parser::parse` instead.
-    fn parse_recovery<Iter: IntoIterator<Item = I>>(&self, iter: Iter) -> (Option<O>, Vec<Self::Error>) where Self: Sized {
+    fn parse_recovery<S: IntoStream<I, <Self::Error as Error<I>>::Span>>(&self, stream: S) -> (Option<O>, Vec<Self::Error>) where Self: Sized {
         let mut errors = Vec::new();
-        match self.parse_inner(&mut IterStream::new(iter.into_iter()), &mut errors).1 {
-            Ok(o) => (Some(o), errors),
+        match self.parse_inner(&mut stream.into_stream(), &mut errors).1 {
+            Ok((o, _)) => (Some(o), errors),
             Err(e) => {
                 errors.push(e);
                 (None, errors)
@@ -140,8 +186,8 @@ pub trait Parser<I, O> {
     /// Parse an iterator of tokens, yielding an output *or* any errors that were encountered along the way.
     ///
     /// If you wish to attempt to produce an output even if errors are encountered, use `Parser::parse_recovery`.
-    fn parse<Iter: IntoIterator<Item = I>>(&self, iter: Iter) -> Result<O, Vec<Self::Error>> where Self: Sized {
-        let (output, errors) = self.parse_recovery(iter);
+    fn parse<S: IntoStream<I, <Self::Error as Error<I>>::Span>>(&self, stream: S) -> Result<O, Vec<Self::Error>> where Self: Sized {
+        let (output, errors) = self.parse_recovery(stream);
         if errors.len() > 0 {
             Err(errors)
         } else {
@@ -149,7 +195,7 @@ pub trait Parser<I, O> {
         }
     }
 
-    /// Map the output of this parser to a value of a potentially different type.
+    /// Map the output of this parser to aanother value.
     ///
     /// # Examples
     ///
@@ -171,11 +217,30 @@ pub trait Parser<I, O> {
     ///
     /// let token = word.or(num);
     ///
-    /// assert_eq!(token.parse("test".chars()), Ok(Token::Word("test".to_string())));
-    /// assert_eq!(token.parse("42".chars()), Ok(Token::Num(42)));
+    /// assert_eq!(token.parse("test"), Ok(Token::Word("test".to_string())));
+    /// assert_eq!(token.parse("42"), Ok(Token::Num(42)));
     /// ```
     fn map<U, F: Fn(O) -> U>(self, f: F) -> Map<Self, F, O> where Self: Sized { Map(self, f, PhantomData) }
 
+    /// Map the output of this parser to another value, making use of the pattern's span.
+    fn map_with_span<U, F: Fn(O, Option<<Self::Error as Error<I>>::Span>) -> U>(self, f: F) -> MapWithSpan<Self, F, O>
+        where Self: Sized
+        { MapWithSpan(self, f, PhantomData) }
+
+    /// Map the primary error of this parser to another value.
+    ///
+    /// This does not map error emitted by sub-patterns within the parser.
+    fn map_err<F: Fn(Self::Error) -> Self::Error>(self, f: F) -> MapErr<Self, F>
+        where Self: Sized
+    { MapErr(self, f) }
+
+    /// Label the pattern parsed by this parser for error messages.
+    ///
+    /// This does not label sub-patterns within the parser.
+    fn label<L: Into<<Self::Error as Error<I>>::Pattern> + Clone>(self, label: L) -> Label<Self, L>
+        where Self: Sized
+    { Label(self, label) }
+
     /// Transform all outputs of this parser to a pretermined value.
     ///
     /// # Examples
@@ -191,11 +256,60 @@ pub trait Parser<I, O> {
     ///     .or(just('*').to(Op::Mul))
     ///     .or(just('/').to(Op::Div));
     ///
-    /// assert_eq!(op.parse("+".chars()), Ok(Op::Add));
-    /// assert_eq!(op.parse("/".chars()), Ok(Op::Div));
+    /// assert_eq!(op.parse("+"), Ok(Op::Add));
+    /// assert_eq!(op.parse("/"), Ok(Op::Div));
     /// ```
     fn to<U: Clone>(self, x: U) -> To<Self, O, U> where Self: Sized { To(self, x, PhantomData) }
 
+    /// Left-fold the output of the parser into a single value, where the output is of type `(_, Vec<_>)`.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use chumsky::prelude::*;
+    ///
+    /// let int = text::int::<Simple<char>>()
+    ///     .collect::<String>()
+    ///     .map(|s| s.parse().unwrap());
+    ///
+    /// let sum = int
+    ///     .then(just('+').padding_for(int).repeated())
+    ///     .foldl(|a, b| a + b);
+    ///
+    /// assert_eq!(sum.parse("1+12+3+9"), Ok(25));
+    /// assert_eq!(sum.parse("6"), Ok(6));
+    /// ```
+    fn foldl<A, B, F: Fn(A, B) -> A>(self, f: F) -> Foldl<Self, F, A, B>
+    where
+        Self: Parser<I, (A, Vec<B>)> + Sized
+    { Foldl(self, f, PhantomData) }
+
+    /// Right-fold the output of the parser into a single value, where the output is of type `(Vec<_>, _)`.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use chumsky::prelude::*;
+    ///
+    /// let int = text::int::<Simple<char>>()
+    ///     .collect::<String>()
+    ///     .map(|s| s.parse().unwrap());
+    ///
+    /// let signed = just('+').to(1)
+    ///     .or(just('-').to(-1))
+    ///     .repeated()
+    ///     .then(int)
+    ///     .foldr(|a, b| a * b);
+    ///
+    /// assert_eq!(signed.parse("3"), Ok(3));
+    /// assert_eq!(signed.parse("-17"), Ok(-17));
+    /// assert_eq!(signed.parse("--+-+-5"), Ok(5));
+    /// ```
+    fn foldr<'a, A, B, F: Fn(A, B) -> B + 'a>(self, f: F) -> Foldr<Self, F, A, B>
+    where
+        Self: Parser<I, (Vec<A>, B)> + Sized
+    { Foldr(self, f, PhantomData) }
+
     /// Ignore the output of this parser, yielding `()` as an output instead.
     ///
     /// This can be used to reduce the cost of passing by avoiding unnecessary allocations (most collections containing
@@ -213,8 +327,8 @@ pub trait Parser<I, O> {
     ///     .ignored()
     ///     .repeated();
     ///
-    /// assert_eq!(whitespace.parse("    ".chars()), Ok(vec![(); 4]));
-    /// assert_eq!(whitespace.parse("  hello".chars()), Ok(vec![(); 2]));
+    /// assert_eq!(whitespace.parse("    "), Ok(vec![(); 4]));
+    /// assert_eq!(whitespace.parse("  hello"), Ok(vec![(); 2]));
     /// ```
     fn ignored(self) -> Ignored<Self, O> where Self: Sized { To(self, (), PhantomData) }
 
@@ -231,7 +345,7 @@ pub trait Parser<I, O> {
     ///     .repeated()
     ///     .collect::<String>();
     ///
-    /// assert_eq!(word.parse("hello".chars()), Ok("hello".to_string()));
+    /// assert_eq!(word.parse("hello"), Ok("hello".to_string()));
     /// ```
     fn collect<C: core::iter::FromIterator<O::Item>>(self) -> Map<Self, fn(O) -> C, O>
         where Self: Sized, O: IntoIterator
@@ -249,11 +363,54 @@ pub trait Parser<I, O> {
     ///     .collect::<String>();
     /// let two_words = word.padded_by(just(' ')).then(word);
     ///
-    /// assert_eq!(two_words.parse("dog cat".chars()), Ok(("dog".to_string(), "cat".to_string())));
-    /// assert!(two_words.parse("hedgehog".chars()).is_err());
+    /// assert_eq!(two_words.parse("dog cat"), Ok(("dog".to_string(), "cat".to_string())));
+    /// assert!(two_words.parse("hedgehog").is_err());
     /// ```
     fn then<U, P: Parser<I, U>>(self, other: P) -> Then<Self, P> where Self: Sized { Then(self, other) }
 
+    /// Parse one thing and then another thing, attempting to chain the two outputs into a [`Vec`].
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use chumsky::prelude::*;
+    ///
+    /// let int = just('-').or_not()
+    ///     .chain(filter::<_, _, Simple<char>>(|c: &char| c.is_ascii_digit() && *c != '0')
+    ///         .chain(filter::<_, _, Simple<char>>(|c: &char| c.is_ascii_digit()).repeated()))
+    ///     .or(just('0').map(|c| vec![c]))
+    ///     .padded_by(end())
+    ///     .collect::<String>()
+    ///     .map(|s| s.parse().unwrap());
+    ///
+    /// assert_eq!(int.parse("0"), Ok(0));
+    /// assert_eq!(int.parse("415"), Ok(415));
+    /// assert_eq!(int.parse("-50"), Ok(-50));
+    /// assert!(int.parse("-0").is_err());
+    /// assert!(int.parse("05").is_err());
+    /// ```
+    fn chain<T, U, P: Parser<I, U, Error = Self::Error>>(self, other: P) -> Map<Then<Self, P>, fn((O, U)) -> Vec<T>, (O, U)>
+    where
+        Self: Sized,
+        U: Chain<T>,
+        O: Chain<T>,
+    {
+        self.then(other).map(|(a, b)| {
+            let mut v = Vec::with_capacity(a.len() + b.len());
+            a.append(&mut v);
+            b.append(&mut v);
+            v
+        })
+    }
+
+    /// Flatten a nested collection.
+    fn flatten<T, Inner>(self) -> Map<Self, fn(O) -> Vec<T>, O>
+    where
+        Self: Sized,
+        O: IntoIterator<Item = Inner>,
+        Inner: IntoIterator<Item = T>,
+    { self.map(|xs| xs.into_iter().map(|xs| xs.into_iter()).flatten().collect()) }
+
     /// Parse one thing and then another thing, yielding only the output of the latter.
     ///
     /// # Examples
@@ -268,8 +425,8 @@ pub trait Parser<I, O> {
     ///     .collect::<String>()
     ///     .map(|s| s.parse().unwrap());
     ///
-    /// assert_eq!(integer.parse("00064".chars()), Ok(64));
-    /// assert_eq!(integer.parse("32".chars()), Ok(32));
+    /// assert_eq!(integer.parse("00064"), Ok(64));
+    /// assert_eq!(integer.parse("32"), Ok(32));
     /// ```
     fn padding_for<U, P: Parser<I, U>>(self, other: P) -> PaddingFor<Self, P, O, U>
         where Self: Sized
@@ -294,7 +451,7 @@ pub trait Parser<I, O> {
     ///     .repeated();
     ///
     /// assert_eq!(
-    ///     sentence.parse("hello! how are you?".chars()),
+    ///     sentence.parse("hello! how are you?"),
     ///     Ok(vec![
     ///         "hello".to_string(),
     ///         "how".to_string(),
@@ -314,7 +471,7 @@ pub trait Parser<I, O> {
     /// If parsing of the inner pattern is successful, the output is `Some(_)`. If an error occurs, the output is
     /// `None`.
     ///
-    /// The delimiters are assymed to allow nesting, so error recovery will attempt to balance the delimiters where
+    /// The delimiters are assumed to allow nesting, so error recovery will attempt to balance the delimiters where
     /// possible. A syntax error within the delimiters should not prevent correct parsing of tokens beyond the
     /// delimiters.
     ///
@@ -336,8 +493,7 @@ pub trait Parser<I, O> {
     ///     .repeated_at_least(1)
     ///     .collect::<String>();
     ///
-    /// let num = filter::<_, _, Simple<char>>(|c: &char| c.is_ascii_digit())
-    ///     .repeated_at_least(1)
+    /// let num = text::int()
     ///     .collect::<String>()
     ///     .map(|s| s.parse().unwrap());
     ///
@@ -351,7 +507,7 @@ pub trait Parser<I, O> {
     ///
     /// // A valid input
     /// assert_eq!(
-    ///     s_expr.parse_recovery("(add (mul 42 3) 15)".chars()),
+    ///     s_expr.parse_recovery("(add (mul 42 3) 15)"),
     ///     (
     ///         Some(SExpr::List(vec![
     ///             SExpr::Ident("add".to_string()),
@@ -369,14 +525,14 @@ pub trait Parser<I, O> {
     /// // An input with a syntax error at position 11! Thankfully, we're able to recover
     /// // and still produce a useful output for later compilation stages (i.e: type-checking).
     /// assert_eq!(
-    ///     s_expr.parse_recovery("(add (mul ! 3) 15)".chars()),
+    ///     s_expr.parse_recovery("(add (mul ! 3) 15)"),
     ///     (
     ///         Some(SExpr::List(vec![
     ///             SExpr::Ident("add".to_string()),
     ///             SExpr::Error,
     ///             SExpr::Num(15),
     ///         ])),
-    ///         vec![Simple::expected_found(11, Some('('), Some('!'))], // A syntax error!
+    ///         vec![Simple::expected_token_found(Some(10..11), vec!['(', '0', ')'], Some('!'))], // A syntax error!
     ///     ),
     /// );
     /// ```
@@ -394,9 +550,9 @@ pub trait Parser<I, O> {
     ///     .or(just('*'))
     ///     .or(just('/'));
     ///
-    /// assert_eq!(op.parse("+".chars()), Ok('+'));
-    /// assert_eq!(op.parse("/".chars()), Ok('/'));
-    /// assert!(op.parse("!".chars()).is_err());
+    /// assert_eq!(op.parse("+"), Ok('+'));
+    /// assert_eq!(op.parse("/"), Ok('/'));
+    /// assert!(op.parse("!").is_err());
     /// ```
     fn or<P: Parser<I, O>>(self, other: P) -> Or<Self, P> where Self: Sized { Or(self, other) }
 
@@ -416,8 +572,8 @@ pub trait Parser<I, O> {
     /// let word_or_question = word
     ///     .then(just('?').or_not());
     ///
-    /// assert_eq!(word_or_question.parse("hello?".chars()), Ok(("hello".to_string(), Some('?'))));
-    /// assert_eq!(word_or_question.parse("wednesday".chars()), Ok(("wednesday".to_string(), None)));
+    /// assert_eq!(word_or_question.parse("hello?"), Ok(("hello".to_string(), Some('?'))));
+    /// assert_eq!(word_or_question.parse("wednesday"), Ok(("wednesday".to_string(), None)));
     /// ```
     fn or_not(self) -> OrNot<Self> where Self: Sized { OrNot(self) }
 
@@ -437,9 +593,9 @@ pub trait Parser<I, O> {
     ///     .map(|s| s.parse().unwrap());
     ///
     /// let sum = num.then(just('+').padding_for(num).repeated())
-    ///     .map(|(head, tail)| tail.into_iter().fold(head, |a, b| a + b));
+    ///     .foldl(|a, b| a + b);
     ///
-    /// assert_eq!(sum.parse("2+13+4+0+5".chars()), Ok(24));
+    /// assert_eq!(sum.parse("2+13+4+0+5"), Ok(24));
     /// ```
     fn repeated(self) -> Repeated<Self> where Self: Sized { Repeated(self, 0) }
 
@@ -456,8 +612,8 @@ pub trait Parser<I, O> {
     ///     .repeated_at_least(5)
     ///     .collect::<String>();
     ///
-    /// assert_eq!(long_word.parse("hello".chars()), Ok("hello".to_string()));
-    /// assert!(long_word.parse("hi".chars()).is_err());
+    /// assert_eq!(long_word.parse("hello"), Ok("hello".to_string()));
+    /// assert!(long_word.parse("hi").is_err());
     /// ```
     fn repeated_at_least(self, n: usize) -> Repeated<Self> where Self: Sized { Repeated(self, n) }
 }
